import math
import torch.nn as nn
from collections import OrderedDict
import ltr.models.target_classifier.linear_filter as target_clf
import ltr.models.target_classifier.features as clf_features
import ltr.models.target_classifier.initializer as clf_initializer
import ltr.models.target_classifier.optimizer as clf_optimizer
import ltr.models.backbone as backbones
from ltr import model_constructor
import torch.nn.functional as F
import torch.distributed as dist
from ltr.models.TransEncoder.transformer import Transformer
from ltr.models.target_classifier.initializer import FilterPool
from ltr.models.TransEncoder.head import build_box_head

class ClsNet_3scales(nn.Module):
    """The FCOTNet network.
    args:
        feature_extractor:  Backbone feature extractor network. Must return a dict of feature maps.
        classification_layer:  Name of the backbone feature layer to use for classification.
        pyramid_*: The blocks in Decoder.
        classifier_18:  Target classification module with feature size of 18.
        classifier_72:  Target classification module with feature size of 72.
        regressor_72:  Anchor-free regression module with feature size of 72.
        train_cls_72_and_reg_init: Whether clssifier_72 and regression initial model should be trained or not.
        train_reg_optimizer:  Whether regression optimizer should be trained or not.
        train_cls_18:  Whether classifier_18 should be trained or not. """

    def __init__(self, feature_extractor, classification_layer, classifier_1, classifier_2, classifier_3, hidden_dim=1024):
        super().__init__()

        self.feature_extractor = feature_extractor
        self.classifier_1 = classifier_1
        self.classifier_2 = classifier_2
        self.classifier_3 = classifier_3
        # self.transformer = classifier_s

        self.proj1 = nn.Linear(256, 1024)
        self.proj2 = nn.Linear(512, 1024)
        self.proj23 = nn.Linear(1024, 1024)
        self.proj123 = nn.Linear(1024, 1024)
        self.dropout23 = nn.Dropout(0.1)
        self.dropout123 = nn.Dropout(0.1)
        self.activation = nn.ReLU()
        # self.proj3 = nn.Linear(1024, hidden_dim)
        # self.target_pool = FilterPool(filter_size=4, feature_stride=16, pool_square=False)
        # self.score_head = build_box_head(hidden_dim, outdim=1)

        self.classification_layer = [classification_layer] if isinstance(classification_layer, str) else classification_layer
        self.output_layers = ['layer1', 'layer2', 'layer3']

        # for func_name in ['classifier_18', 'feature_extractor']:
        #     for p in getattr(self, func_name).parameters():
        #         p.requires_grad_(False)
        # for func_name in ['classifier_72', 'regressor_72', 'pyramid_first_conv', 'pyramid_36', 'pyramid_72']:
        #     for p in getattr(self, func_name).parameters():
        #         p.requires_grad_(False)

    def forward(self, train_imgs, test_imgs, train_bb, *args, **kwargs):

        assert train_imgs.dim() == 5 and test_imgs.dim() == 5, 'Expect 5 dimensional inputs'
        n, b, _, _, _ = test_imgs.size()

        # Extract backbone features
        train_feat = self.extract_backbone_features(train_imgs.view(-1, *train_imgs.shape[-3:]),
                                                    layers=['layer1','layer2','layer3'])
        test_feat = self.extract_backbone_features(test_imgs.view(-1, *test_imgs.shape[-3:]),
                                                   layers=['layer1','layer2','layer3'])

        # Classification features
        # train_feat_s = self.proj3(train_feat['layer3'].permute(0, 2, 3, 1)).permute(0, 3, 1, 2)
        # test_feat_s = self.proj3(test_feat['layer3'].permute(0, 2, 3, 1)).permute(0, 3, 1, 2)

        train_feat_3, train_feat_23, train_feat_123 = self.extract_3scales_feat(train_feat)
        test_feat_3, test_feat_23, test_feat_123 = self.extract_3scales_feat(test_feat)

        # train_feat_3 = train_feat['layer3']
        # test_feat_3 = test_feat['layer3']
        #
        # train_feat_2 = F.interpolate(self.proj2(train_feat['layer2'].permute(0, 2, 3, 1)).permute(0, 3, 1, 2), size=(18, 18))
        # test_feat_2 = F.interpolate(self.proj2(test_feat['layer2'].permute(0, 2, 3, 1)).permute(0, 3, 1, 2), size=(18, 18))
        #
        # train_feat_1 = F.interpolate(self.proj1(train_feat['layer1'].permute(0, 2, 3, 1)).permute(0, 3, 1, 2), size=(18, 18))
        # test_feat_1 = F.interpolate(self.proj1(test_feat['layer1'].permute(0, 2, 3, 1)).permute(0, 3, 1, 2), size=(18, 18))
        #
        # train_feat_23 = self.dropout23(self.activation(self.proj23((train_feat_2 + train_feat_3).permute(0, 2, 3, 1)))).permute(0, 3, 1, 2)
        # train_feat_123 = self.dropout123(self.activation(self.proj123((train_feat_2 + train_feat_1 + train_feat_3).permute(0, 2, 3, 1)))).permute(0, 3, 1, 2)
        #
        # test_feat_23 = self.dropout23(self.activation(self.proj23((test_feat_2 + test_feat_3).permute(0, 2, 3, 1)))).permute(0, 3, 1, 2)
        # test_feat_123 = self.dropout123(self.activation(self.proj123((test_feat_2 + test_feat_1 + test_feat_3).permute(0, 2, 3, 1)))).permute(0, 3, 1, 2)

        # generate score_map
        # target0 = train_feat_s.view(n, b, *train_feat_s.shape[-3:])[0]
        # target = self.generate_target(target0, train_bb[0])  # B, C, Ht, Wt
        # target = target.unsqueeze(0).repeat(n, 1, 1, 1, 1).view(n*b, -1, 4, 4)
        # target = target.flatten(2).permute(2, 0, 1)
        # search = test_feat_s.flatten(start_dim=2).permute(2, 0, 1)  # HWxNBxC
        # hs = self.transformer(search, target)  # HWxNBxC
        # score_s = self.score_head(hs).squeeze().permute(1, 0).view(n, b, *test_feat_s.shape[-2:])

        scores_3 = self.classifier_3(train_feat_3, test_feat_3, train_bb, *args, **kwargs)
        # scores_2 = self.classifier_2(train_feat_2, test_feat_2, train_bb, *args, **kwargs)
        # scores_1 = self.classifier_1(train_feat_1, test_feat_1, train_bb, *args, **kwargs)
        scores_2 = self.classifier_2(train_feat_23, test_feat_23, train_bb, *args, **kwargs)
        scores_1 = self.classifier_1(train_feat_123, test_feat_123, train_bb, *args, **kwargs)

        output = {}
        # output['scores_s'] = score_s
        output['scores_1'] = scores_1
        output['scores_2'] = scores_2
        output['scores_3'] = scores_3

        return output

    def extract_3scales_feat(self, feat):
        feat_3 = feat['layer3']
        feat_2 = F.interpolate(self.proj2(feat['layer2'].permute(0, 2, 3, 1)).permute(0, 3, 1, 2), size=(18, 18))
        feat_1 = F.interpolate(self.proj1(feat['layer1'].permute(0, 2, 3, 1)).permute(0, 3, 1, 2), size=(18, 18))

        feat_23 = self.dropout23(self.activation(self.proj23((feat_2 + feat_3).permute(0, 2, 3, 1)))).permute(0, 3, 1, 2)
        feat_123 = self.dropout123(self.activation(self.proj123((feat_2 + feat_1 + feat_3).permute(0, 2, 3, 1)))).permute(0, 3, 1, 2)
        return feat_3, feat_23, feat_123

    def generate_target(self, feat_train, bb):
        bb_pool = bb.clone()
        # print("bb: {}".format(bb))
        target = self.target_pool(feat_train, bb_pool)

        # target = target.view(-1, self.filter_dim, target.shape[-3] // self.filter_dim, target.shape[-2], target.shape[-1])
        return target


    def get_backbone_clf_feat(self, backbone_feat):
        feat = OrderedDict({l: backbone_feat[l] for l in self.classification_layer})
        if len(self.classification_layer) == 1:
            return feat[self.classification_layer[0]]
        return feat

    def extract_backbone_features(self, im, layers=None):
        if layers is None:
            layers = self.output_layers
        return self.feature_extractor(im, layers)

    def extract_features(self, im, layers=None):
        if layers is None:
            layers = self.bb_regressor_layer + ['classification']
        if 'classification' not in layers:
            return self.feature_extractor(im, layers)
        backbone_layers = sorted(list(set([l for l in layers + self.classification_layer if l != 'classification'])))
        all_feat = self.feature_extractor(im, backbone_layers)
        all_feat['classification'] = self.extract_classification_feat(all_feat)
        return OrderedDict({l: all_feat[l] for l in layers})

def is_dist_avail_and_initialized():
    if not dist.is_available():
        return False
    if not dist.is_initialized():
        return False
    return True

def get_rank():
    if not is_dist_avail_and_initialized():
        return 0
    return dist.get_rank()

def is_main_process():
    return get_rank() == 0

@model_constructor
def clsnet_3scales(clf_filter_size=4, optim_iter=5, optim_init_step=1.0, optim_init_reg=0.01,
            classification_layer='layer3', feat_stride=16, clf_feat_blocks=0,
            clf_feat_norm=True, init_filter_norm=False, final_conv=True, out_feature_dim=512,
            init_gauss_sigma=1.0, num_dist_bins=5, bin_displacement=1.0, mask_init_factor=4.0,
            score_act='relu', act_param=None, target_mask_act='sigmoid', detach_length=float('Inf'),
            frozen_backbone_layers=(), freeze_bn=True):
    # Backbone
    backbone_net = backbones.resnet50(pretrained=is_main_process(), frozen_layers=frozen_backbone_layers,
                                      freeze_bn=freeze_bn)
    # static classifier (layer1-18)
    # transformer = Transformer(d_model=256, nhead=8, num_encoder_layers=6, num_decoder_layers=6)

    # classifier_3 (We use the same architecture of classifier_18 with DiMP.)
    norm_scale_3 = math.sqrt(1.0 / (out_feature_dim * clf_filter_size * clf_filter_size))
    clf_head_3 = clf_features.clf_head_18(num_blocks=clf_feat_blocks, l2norm=clf_feat_norm,
                                           final_conv=final_conv, norm_scale=norm_scale_3,
                                           out_dim=out_feature_dim, feature_dim=256)  # , feature_dim=256
    initializer_3 = clf_initializer.FilterInitializerLinear(filter_size=clf_filter_size, filter_norm=init_filter_norm,
                                                             feature_dim=out_feature_dim)
    optimizer_3 = clf_optimizer.DiMPSteepestDescentGN(num_iter=optim_iter, feat_stride=feat_stride,
                                                       init_step_length=optim_init_step,
                                                       init_filter_reg=optim_init_reg,
                                                       init_gauss_sigma=init_gauss_sigma,
                                                       num_dist_bins=num_dist_bins,
                                                       bin_displacement=bin_displacement,
                                                       mask_init_factor=mask_init_factor,
                                                       score_act=score_act, act_param=act_param,
                                                       mask_act=target_mask_act,
                                                       detach_length=detach_length)
    classifier_3 = target_clf.LinearFilter(filter_size=clf_filter_size, filter_initializer=initializer_3,
                                            filter_optimizer=optimizer_3, feature_extractor=clf_head_3)

    # classifier_2 (We use the same architecture of classifier_18 with DiMP.)
    norm_scale_2 = math.sqrt(1.0 / (out_feature_dim * clf_filter_size * clf_filter_size))
    clf_head_2 = clf_features.clf_head_18(num_blocks=clf_feat_blocks, l2norm=clf_feat_norm,
                                          final_conv=final_conv, norm_scale=norm_scale_2,
                                          out_dim=out_feature_dim, feature_dim=256)
    initializer_2 = clf_initializer.FilterInitializerLinear(filter_size=clf_filter_size, filter_norm=init_filter_norm,
                                                            feature_dim=out_feature_dim)
    optimizer_2 = clf_optimizer.DiMPSteepestDescentGN(num_iter=optim_iter, feat_stride=feat_stride,
                                                      init_step_length=optim_init_step,
                                                      init_filter_reg=optim_init_reg,
                                                      init_gauss_sigma=init_gauss_sigma,
                                                      num_dist_bins=num_dist_bins,
                                                      bin_displacement=bin_displacement,
                                                      mask_init_factor=mask_init_factor,
                                                      score_act=score_act, act_param=act_param,
                                                      mask_act=target_mask_act,
                                                      detach_length=detach_length)
    classifier_2 = target_clf.LinearFilter(filter_size=clf_filter_size, filter_initializer=initializer_2,
                                           filter_optimizer=optimizer_2, feature_extractor=clf_head_2)

    # classifier_1 (We use the same architecture of classifier_18 with DiMP.)
    norm_scale_1 = math.sqrt(1.0 / (out_feature_dim * clf_filter_size * clf_filter_size))
    clf_head_1 = clf_features.clf_head_18(num_blocks=clf_feat_blocks, l2norm=clf_feat_norm,
                                          final_conv=final_conv, norm_scale=norm_scale_1,
                                          out_dim=out_feature_dim, feature_dim=256)  # , feature_dim=64
    initializer_1 = clf_initializer.FilterInitializerLinear(filter_size=clf_filter_size, filter_norm=init_filter_norm,
                                                            feature_dim=out_feature_dim)
    optimizer_1 = clf_optimizer.DiMPSteepestDescentGN(num_iter=optim_iter, feat_stride=feat_stride,
                                                      init_step_length=optim_init_step,
                                                      init_filter_reg=optim_init_reg,
                                                      init_gauss_sigma=init_gauss_sigma,
                                                      num_dist_bins=num_dist_bins,
                                                      bin_displacement=bin_displacement,
                                                      mask_init_factor=mask_init_factor,
                                                      score_act=score_act, act_param=act_param,
                                                      mask_act=target_mask_act,
                                                      detach_length=detach_length)
    classifier_1 = target_clf.LinearFilter(filter_size=clf_filter_size, filter_initializer=initializer_1,
                                           filter_optimizer=optimizer_1, feature_extractor=clf_head_1)

    # FCOT network
    net = ClsNet_3scales(feature_extractor=backbone_net, classification_layer=classification_layer,
                  classifier_1=classifier_1, classifier_2=classifier_2,
                  classifier_3=classifier_3)
    return net